/**
 * Host representation
 * 
 * @param {Integer} hostId Identifier or the link on the network
 * @param {Object} link Link representation on which host is connected
 * @returns Public methods for a host
 */
function Host(hostId, link) {
	// Global host variables
	var _c = 2.3 * Math.pow (10,8); // Light speed in copper 
	var _outbox = $('#host-'+hostId+' ul');
	var SENSE_DURATION = 0.000096;
	var _packetsPerFrame = null; // Will be either a number (audio or video) or list (poisson process)
	var MAX_PACKET_PER_SEC = 1 / SENSE_DURATION;
		
	// Host properties
	var _properties = {
		id: hostId,
		packetPayload: 0,
		hostDistance: 0, // Consecutive host distance
		errorRate: 0, // Percents of error when sending a packet
		bandwith: 0, // Bandwidth between links in bits/s
		dataSize: 0, // Size of the data to send
		sentPackets: 0, // Number of packets sent
		sentFrames: 0, // Number of frame sent
		numPackets: 0, // Number of packets sent
		receiverId: -1,
		distance: 0, // Distance to the current receiver
		hosts: [], // Host on the networks
		linkIdle: true,
		jam: false, // JAM signal receive
		sendingUntil: 0, // Time until where are sending
		lastPacketIds: [],
		resentLast: false,
		connected: false, // Do we are in connected mode (reliable)
		backoffCoeff: 0,
		jamSignal: 0, // Consecutive JAM signal
		dataType: '', // Can be 'audio', 'video' or 'data'
		collisionOnPrevious: false,
	};
	
	/**
	 * Sense network before sending a packet
	 * 
	 * @param {Integer} duration Optional sensing duration. Assume if specified
	 * 							 that a collision has occurred.
	 */
	var _sense = function (duration) {
		if (_properties.sentPackets >= _properties.numPackets) {
			return;
		}
		
		_properties.linkIdle = true;

		if (!duration) {
			if (_properties.collisionOnPrevious) {
				_properties.collisionOnPrevious = false;
			} else {
				_properties.backoffCoeff = 0;
				_properties.jamSignal = 0;
			}
			
			if (_properties.sendingUntil == 0) {
				_properties.sendingUntil = CLOCK;
			}
			
			duration = _properties.sendingUntil + SENSE_DURATION;
		} else {
			_properties.collisionOnPrevious = true;
		}
		
		link.sense(_properties.hosts[_properties.id-1], duration);
	};
	
	/**
	 * Send a packet to a the defined receiver
	 */
	var _send = function () {
		if (_properties.linkIdle) {
			var numPacketToSend = _frameSize(_properties.dataType, _properties.sentFrames);
			var packets = [];
			var times = [];
			var sentIds = [];
			
			_properties.sendingUntil = CLOCK;
			
			// Don't need to keep packet ID in connection-less
			if (!_properties.connected) {
				_properties.lastPacketIds = [];
			}
			
			for (var i=0; i < numPacketToSend; i++) {
				var packet = Packet();
				packet.create(_properties.hosts[_properties.id-1],
							  _properties.hosts[_properties.receiverId], 
							  _properties.packetPayload, 
							  _properties.errorRate,
							  !((!_properties.resentLast) || _properties.lastPacketIds.pop()));
				
				_properties.lastPacketIds.push(packet.get('id'));
				_properties.sendingUntil += _delay();
				_properties.sentPackets++;

				sentIds.push(packet.get('id'));
				
				packets.push(packet);
				times.push(_properties.sendingUntil);

				Stats.sent(_properties.id, numPacketToSend*_properties.packetPayload);
			}

			_output('Sending packet #' + sentIds.join(', #') + ' to host ' + (_properties.receiverId+1), 'info');
			_properties.resentLast = false;
			_properties.sentFrames++;

			link.send(packets, times);
			
			
			// Prepare next sensing
			_sense();
		} else {
			_output('Defering sending, link in use', 'notice');
			_sense();
		}
	};
	
	/**
	 * Define the number of packet to send in the frame for a given packetId
	 * 
	 * @param {Integer} packetId
	 */
	var _frameSize = function (type, frameId) {
		var bitrate = 0;
		if (type == 'data') {
			if (!_packetsPerFrame) {
				_packetsPerFrame = _poissonProcess();
			}

			return _packetsPerFrame[frameId];
		} else if (type == 'video') {
			bitrate = 1572864; // 1.5Mpbs
		} else {
			bitrate = 65536; // 64 Kbps
		}
		
		if (!_packetsPerFrame) {
			var packetPerSec = Math.ceil(bitrate / _properties.packetPayload);
			_packetsPerFrame = Math.ceil(packetPerSec / MAX_PACKET_PER_SEC);
		}
		
		return _packetsPerFrame;
	};
	
	/**
	 * Compute poisson process
	 */
	var _poissonProcess = function () {
		var poissonDistribution = [];
		var poissonProcess = [];
		var lambda = _properties.numPackets / 2;
		var frameSizeProba;
		
		for (var i=0; i < _properties.numPackets; i++) {
			var exp = Math.exp(-lambda);
			var denominator = Math.pow(lambda, i);
			var numerator = _factorial(i);
			poissonDistribution[i] = Math.ceil(exp * (denominator / numerator) * _properties.numPackets);
		}
		
		frameSizeProba = poissonDistribution[0];
		for (var i=0; frameSizeProba < _properties.numPackets; i++) {
			if (i > 0) {
				frameSizeProba += poissonProcess[i-1];
			}
			
			poissonProcess[i] = poissonDistribution[frameSizeProba-1];
 		}
		
		return poissonProcess;
	};
	
	/**
	 * Factorial computation
	 * 
	 * @param {Integer} n
	 */
	var _factorial = function (n) {
		return (n === 0) ? 1 : n*_factorial(n-1);
	};
	
	/**
	 * Delay to send a packet between hosts
	 * 
	 * @returns {Integer}  Transmit delay
	 */
	var _delay = function (){
		var propagationTime = _properties.distance / _c;
		var transmitTime = _properties.packetPayload.length / _properties.bandwith;
		return propagationTime + transmitTime;
	};
	
	/**
	 * Receive a packet from a sender
	 * 
	 * @param packet Received packet
	 */
	var _receive = function(packet) {
		var haveError = (packet.crc() == packet.get('checkseq')) ? 'success' : 'error';
		
		_output('Received packet #' + packet.get('id') + ' from host ' + packet.get('source').get('id'), haveError);
		
		Stats.received(_properties.id, packet.get('payload').length, (haveError == 'error'));
		
		// TODO: Use propagation delay
		if (_properties.connected) {
			packet.get('source').ack(packet.get('id'));
		}
	};
	
	/**
	 * Acknowledge packet reception
	 * 
	 *  @param {Integer|Object} packetId
	 */
	var _ack = function (packetId) {
		_properties.lastPacketIds.remove(packetId);
	};
	
	/**
	 * JAM signal handler
	 */
	var _jam = function () {
		_properties.jamSignal++;
		_properties.backoffCoeff++;
		
		// Limit backoff to 10
		if (_properties.backoffCoeff > 10) {
			_properties.backoffCoeff = 10;
		}
		
		if ((_properties.sendingUntil > 0) && (_properties.sendingUntil >= CLOCK)) {

			if (_properties.connected) {
				// Retry 15 time max
				if (_properties.jamSignal < 15) {
					_properties.sentPackets -= _properties.lastPacketId.length;
					_properties.sentFrames--;
					_properties.resentLast = true;
				} else {
					_output('Aborting sending packet #' + _properties.lastPacketIds.join(', #'));
				}
				
				_output('Resending collided-packet #' + _properties.lastPacketIds.join(', #') + ', retry #' + _properties.backoffCoeff, 'error');
				
			} else {
				_output('Packet #' + _properties.lastPacketIds.join(', #') + ' has collided, retry #' + _properties.backoffCoeff, 'error');
			}
		} else {
			_output('Collision detected', 'notice');
		}

		_properties.sendingUntil = 0;
		_properties.linkIdle = true;
		
		var sensingTime = CLOCK + SENSE_DURATION + ((Math.random() * Math.pow(2, _properties.backoffCoeff)) * 0.000512);
		
		_sense(sensingTime);
	};

	/**
	 * Define an action to be done by the host.
	 * 
	 * @param {String} actionType
	 * @param {Integer} hostId On which host the action is performed
	 * @param {Integer} avgDataSize Average size of the packet payload
	 */
	var _setAction = function (actionType, hostId, avgDataSize) {
		var sizeVariance = 0.5 * avgDataSize;
		_properties.dataSize = avgDataSize + Math.ceil((Math.random() *  sizeVariance) - (sizeVariance/2));
		_properties.numPackets = Math.ceil(_properties.dataSize / _properties.packetPayload);
		_properties.receiverId = hostId;
		_properties.distance = Math.abs(_properties.id - hostId) * _properties.hostDistance;
		_properties.dataType = actionType;
		
		_output('Will send  '+ _bytesToSize(_properties.dataSize) + ' ('+ _properties.numPackets +' packets) of ' + actionType + ' to host ' + (hostId+1));
		
		// Start sensing to send, but avoid every body sense the same time at start
		// to avoid unrealistic collisions 
		_sense(SENSE_DURATION + (Math.random() * SENSE_DURATION));
	};
	
	/**
	 * Create human readable size
	 * 
	 * @param {Double} bytes
	 */
	var _bytesToSize = function (bytes) {
	    var sizes = ['B', 'KB', 'MB', 'GB', 'TB'];
	    if (bytes == 0) return 'n/a';
	    var i = parseInt(Math.floor(Math.log(bytes) / Math.log(1024)));
	    return (bytes / Math.pow(1024, i)).toFixed(1) + ' ' + sizes[i];
	};
	
	/**
	 * Set host property
	 * 
	 * @param {String} propertyName
	 * @param {Object} value
	 */
	var _set = function (propertyName, value) {
		_properties[propertyName] = value;
	}; 
	
	/**
	 * Get host property
	 * script
	 * @param {String} propertyName
	 */
	var _get = function (propertyName) {
		return _properties[propertyName];
	};
	
	/**
	 * Crop decimal to a precision
	 * 
	 * @param {Double} num Number to work on
	 * @param {Integer} precision
	 */
	var _roundDecimal = function (num, precision) {
		var result = Math.round(num*Math.pow(10, precision))/Math.pow(10, precision);
		return result;
	};
	
	/**
	 * Write output in the host box
	 * 
	 * @param {String} message
	 * @param {String} status Class of the message (error, notice, success, info)
	 */
	var _output = function (message, status) {
		var stat = status ? ' class="'+status+'"' : '';
		_outbox.append('<li'+stat+'>'+_roundDecimal(CLOCK, 10)+': '+message+'</li>');
	};
	
	/**
	 * Initialize host
	 * 
	 * @param {Object} settings Host properties
	 */
	var _init = function (settings) {
		for (attr in settings) {
			_properties[attr] = settings[attr];
		}		
	};
	
	// Expose methods
	return {
		sense: _sense,
		send: _send,
		jam: _jam,
		receive: _receive,
		ack: _ack,
		get: _get,
		set: _set,
		setAction: _setAction,
		init: _init,
	};
}
